-----------------------------------------------------------------------------------------------
-- Client Lua Script for Defrag
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
require "Window"

-----------------------------------------------------------------------------------------------
-- Defrag Module Definition
-----------------------------------------------------------------------------------------------
local Defrag = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local CodeVersion = "1.6"

local sin = math.sin
local cos =  math.cos
local rad = math.rad

function Defrag:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
    return o
end

function Defrag:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Defrag"
	local tDependencies = { }
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

local DefragInst = Defrag:new()
DefragInst:Init()

function Defrag:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("Defrag.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function Defrag:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.Overlay = Apollo.LoadForm(self.xmlDoc, "DefragForm","InWorldHudStratum", self)

		if self.Overlay == nil then
			Apollo.AddAddonErrorText(self, "Could not load the  window for some reason.")
			return
		end
		
	    self.Overlay:Show(true, true)
		
		if (self.tSettings == nil) then
		end
		
		--initialize variables
		self.bEnabled = false
		self.tDefragCount = {} 
		self.tMazeCount = {} 
		self.timerDefrag = nil
		self.timerDefragFails = nil
		self.timerMazeFails = nil
		self.timerHitBox = nil
		self.uWater = nil
		self.uLogic = nil
		self.uWaterColor = "6600ff00"
		self.uLogicColor = "6600ff00"

		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
		Apollo.RegisterEventHandler("_LCLF_SpellCastStart", "OnCastStart", self)
	end
end

function Defrag:OnEnteredCombat(unit, combat)
	if (unit:GetName() == "Hydroflux") or (unit:GetName() == "Mnemesis") then
		if (combat) and (self.timerHitBox == nil) then
			Apollo.RegisterEventHandler("NextFrame", "OnNextFrame", self)
			self.timerHitBox = ApolloTimer.Create(0.1, true, "OnBuffCheck", self)
		elseif (combat ~= true) and (self.timerHitBox ~= nil) then
			self.Overlay:DestroyAllPixies()
			self.timerHitBox:Stop()
			self.timerHitBox = nil
			Apollo.RemoveEventHandler("NextFrame", self)
		end
	end
	
	if (unit:GetName() == "Hydroflux") then
		self.uWater = unit
	end
	
	if (unit:GetName() == "Mnemesis") then
		self.uLogic = unit
	end
end

function Defrag:OnCastStart(strSpellName, tCasterUnit)
	if (tCasterUnit:GetName() == "Mnemesis") and (strSpellName == "Defragment") then
		self.bEnabled = true
		self.tDefragCount = {} 
		Apollo.RegisterEventHandler("CombatLogDamage","OnCombatLogDamage", self)
		self.timerDefrag      = ApolloTimer.Create(9, true, "OnDefragEnd", self)
		self.timerDefragFails = ApolloTimer.Create(10, true, "CountDefragFails", self)
		self.timerDefrag:Start()
		self.timerDefragFails:Start()		
	end
	
	if (tCasterUnit:GetName() == "Mnemesis") and (strSpellName == "Circuit Breaker") then
		self.tMazeCount = {} 
		Apollo.RegisterEventHandler("CombatLogDamage","OnCombatLogDamage", self)
		self.timerMazeFails = ApolloTimer.Create(26, true, "CountMazeFails", self)
	end
end

function Defrag:OnBuffCheck()
	if (self.uWater ~= nil) then
		if (self.uWater:GetBuffs() ~= nil) then
			local tBuffs = self.uWater:GetBuffs().arBeneficial
			local count = 0
			for idx, buff in pairs(tBuffs) do
				if (buff.splEffect:GetName() == "Elemental Alignment") then
					count = count + 1
				end
			end
			if (count == 2) then
				self.uWaterColor = "66ff0000"
			else
				self.uWaterColor = "6600ff00"
			end
		end
	end
	
	if (self.uLogic ~= nil) then
		if (self.uLogic:GetBuffs() ~= nil) then
			local tBuffs = self.uLogic:GetBuffs().arBeneficial
			local count = 0
			for idx, buff in pairs(tBuffs) do
				if (buff.splEffect:GetName() == "Elemental Alignment") then
					count = count + 1
				end
			end
			if (count == 2) then
				self.uLogicColor = "66ff0000"
			else
				self.uLogicColor = "6600ff00"
			end
		end
	end
end

function Defrag:OnDefragEnd()
	self.bEnabled = false
	self.Overlay:DestroyAllPixies()
	self.timerDefrag:Stop()
	self.timerDefrag = nil
end

function Defrag:CountDefragFails()
	if (self.timerMazeFails == nil) then
		Apollo.RemoveEventHandler("CombatLogDamage", self)
		Apollo.RemoveEventHandler("AttackMissed", self)
	end
	self.timerDefragFails:Stop()
	self.timerDefragFails = nil
	
	self:PrintFails(self.tDefragCount, "Defrag")
	self.tDefragCount = {}
end

function Defrag:CountMazeFails()
	Apollo.RemoveEventHandler("CombatLogDamage", self)
	Apollo.RemoveEventHandler("AttackMissed", self)		
	self.timerMazeFails:Stop()
	self.timerMazeFails = nil
	
	self:PrintFails(self.tMazeCount, "Maze")
	self.tMazeCount = {} 
end

function Defrag:OnNextFrame()
	self.Overlay:DestroyAllPixies()
	
	local count, iPos, uFace, iVec, uAngle
	
	if (self.uWater ~= nil) then
		iPos = self.uWater:GetPosition()
		uFace = self.uWater:GetFacing()
		if (iPos ~= nil) and (uFace ~= nil) then
			iVec = Vector3.New(iPos.x, iPos.y, iPos.z)
			uAngle = math.atan2(uFace.x, uFace.z)
			self:PlotCircle(iVec, 5, self.uWaterColor, uAngle)
		end
	end
	
	if (self.uLogic ~= nil) then
		iPos = self.uLogic:GetPosition()
		uFace = self.uLogic:GetFacing()
		if (iPos ~= nil) and (uFace ~= nil) then
			iVec = Vector3.New(iPos.x, iPos.y, iPos.z)
			uAngle = math.atan2(uFace.x, uFace.z)
			self:PlotCircle(iVec, 4, self.uLogicColor, uAngle)
		end	
	end

	if (self.bEnabled == false) or (GameLib.GetPlayerUnit():IsDead()) then return end

	count = 0
	iPos = GameLib.GetPlayerUnit():GetPosition()
	uFace = GameLib.GetPlayerUnit():GetFacing()
	if (iPos ~= nil) and (uFace ~= nil) then
		iVec = Vector3.New(iPos.x, iPos.y, iPos.z)
		uAngle = math.atan2(uFace.x, uFace.z)
		self:PlotSquare(iVec, 13, "ffffffff", uAngle)
		
		for i = 1, GroupLib.GetMemberCount(), 1 do
			local gmember = GroupLib.GetGroupMember(i)
			local member = GroupLib.GetUnitForGroupMember(i)
			if (member) and (count <= 4) then
				if (member ~= GameLib.GetPlayerUnit()) and (member:IsDead() ~= true) then
					mPos = member:GetPosition()
					mVec = Vector3.New(mPos.x, mPos.y, mPos.z)
					dPos = (iVec - mVec)
		
					-- Only draw/update 4 units within 13m, currently do not care which since everyone should spread out
					sqrDist = math.pow(dPos.x, 2) + math.pow(dPos.y, 2) + math.pow(dPos.z, 2)
					if (sqrDist <= 169) then
						count = count + 1
						uFace = member:GetFacing()
						uAngle = math.atan2(uFace.x, uFace.z)
						self:PlotSquare(mVec, 13, "ffff0000", uAngle)
					end
				end	
			end
		end
	end			
end

-- Borrowed from BuffTelegraph, written by Kurona Silversky. Draws a perfect square at resolution '4' and handles unit facing.
function Defrag:PlotSquare(iVec, dist, color, uangle)
	local pVec = Vector3.New(iVec.x + dist * sin(uangle), iVec.y, iVec.z + dist * cos(uangle))
	local startPoint = GameLib.WorldLocToScreenPoint(pVec)
	local pPoints = {0,0,0,0}
	
	local rez = 4
	
	for i = 1, rez do 
		local rot = uangle + rad(-360 / (rez/(i)))

		local curVec = Vector3.New(iVec.x + dist * sin(rot), iVec.y, iVec.z + dist * cos(rot))
		local endPoint = GameLib.WorldLocToScreenPoint(curVec)

		self.Overlay:AddPixie( { bLine = true, fWidth = 4,  cr = color, loc = { fPoints = pPoints, nOffsets = { startPoint.x, startPoint.y, endPoint.x, endPoint.y } } } )

		pVec = curVec
		startPoint = endPoint
	end
end

function Defrag:PlotCircle(iVec, dist, color, uangle)
	local pVec = Vector3.New(iVec.x + dist * sin(uangle), iVec.y, iVec.z + dist * cos(uangle))
	local startPoint = GameLib.WorldLocToScreenPoint(pVec)
	local pPoints = {0,0,0,0}
	
	local rez = 20
	
	for i = 1, rez do 
		local rot = uangle + rad(-360 / (rez/(i)))

		local curVec = Vector3.New(iVec.x + dist * sin(rot), iVec.y, iVec.z + dist * cos(rot))
		local endPoint = GameLib.WorldLocToScreenPoint(curVec)

		self.Overlay:AddPixie( { bLine = true, fWidth = 4,  cr = color, loc = { fPoints = pPoints, nOffsets = { startPoint.x, startPoint.y, endPoint.x, endPoint.y } } } )

		pVec = curVec
		startPoint = endPoint
	end
end

function Defrag:OnCombatLogDamage(tEventArgs)
	if (tEventArgs.strSpell == "Defragment") then
		local unitName = tEventArgs.unitTarget:GetName()
		if (self.tDefragCount[unitName] ~= nil) then
			self.tDefragCount[unitName] = self.tDefragCount[unitName] + 1
		else
			self.tDefragCount[unitName] = 1
		end
	end
	
	if (tEventArgs.strSpell == "Combinatorial Genomics") then
		local unitName = tEventArgs.unitTarget:GetName()
		if (self.tMazeCount[unitName] ~= nil) then
			self.tMazeCount[unitName] = self.tMazeCount[unitName] + 1
		else
			self.tMazeCount[unitName] = 1
		end
	end
end

function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function Defrag:PrintFails(tCounts, strName)
	local count = 0
	local str = strName .. " Fails: "
	for i, v in spairs(tCounts, function(t,a,b) return t[a] > t[b] end) do
		if (v >= 2) then
			count = count + 1
			if (count == 1) then
				str = str .. i .. " (" .. v .. ")"
			else
				str = str .. ", " .. i .. " (" .. v .. ")"
			end
		end
	end
	
	if (count > 0) then
		Print(str)
	end
end


